=== modified file 'bzrlib/__init__.py'
--- bzrlib/__init__.py	2005-08-24 05:51:34 +0000
+++ bzrlib/__init__.py	2005-08-25 06:14:17 +0000
@@ -16,10 +16,6 @@
 
 """bzr library"""
 
-# TODO: Do less imports here
-from branch import Branch, ScratchBranch, find_branch
-from errors import BzrError
-
 BZRDIR = ".bzr"
 
 DEFAULT_IGNORE = ['.bzr.log',
@@ -47,6 +43,9 @@
 
 def get_bzr_revision():
     """If bzr is run from a branch, return (revno,revid) or None"""
+    import bzrlib.errors
+    from bzrlib.branch import Branch
+    
     try:
         branch = Branch(__path__[0])
         rh = branch.revision_history()
@@ -54,7 +53,7 @@
             return len(rh), rh[-1]
         else:
             return None
-    except BzrError:
+    except bzrlib.errors.BzrError:
         return None
     
 def test_suite():

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-08-25 06:08:36 +0000
+++ bzrlib/branch.py	2005-08-25 06:14:17 +0000
@@ -34,6 +34,7 @@
 import bzrlib.ui
 
 
+
 BZR_BRANCH_FORMAT = "Bazaar-NG branch, format 0.0.4\n"
 ## TODO: Maybe include checks for common corruption of newlines, etc?
 
@@ -809,51 +810,46 @@
 
     def update_revisions(self, other, stop_revision=None):
         """Pull in all new revisions from other branch.
-        
-        >>> from bzrlib.commit import commit
-        >>> bzrlib.trace.silent = True
-        >>> br1 = ScratchBranch(files=['foo', 'bar'])
-        >>> br1.add('foo')
-        >>> br1.add('bar')
-        >>> commit(br1, "lala!", rev_id="REVISION-ID-1", verbose=False)
-        >>> br2 = ScratchBranch()
-        >>> br2.update_revisions(br1)
-        Added 2 texts.
-        Added 1 inventories.
-        Added 1 revisions.
-        >>> br2.revision_history()
-        [u'REVISION-ID-1']
-        >>> br2.update_revisions(br1)
-        Added 0 texts.
-        Added 0 inventories.
-        Added 0 revisions.
-        >>> br1.text_store.total_size() == br2.text_store.total_size()
-        True
         """
-        progress = bzrlib.ui.ui_factory.progress_bar()
-        progress.update('comparing histories')
+        from bzrlib.fetch import greedy_fetch
+
+        pb = bzrlib.ui.ui_factory.progress_bar()
+        pb.update('comparing histories')
+
         revision_ids = self.missing_revisions(other, stop_revision)
-        count = self.install_revisions(other, revision_ids, progress=progress)
+
+        if len(revision_ids) > 0:
+            count = greedy_fetch(self, other, revision_ids[-1], pb)[0]
+        else:
+            count = 0
         self.append_revision(*revision_ids)
-        print "Added %d revisions." % count
-                    
+        ## note("Added %d revisions." % count)
 
-    def install_revisions(self, other, revision_ids, progress=None):
+        
+    def install_revisions(self, other, revision_ids, pb):
         if hasattr(other.revision_store, "prefetch"):
             other.revision_store.prefetch(revision_ids)
         if hasattr(other.inventory_store, "prefetch"):
             inventory_ids = [other.get_revision(r).inventory_id
                              for r in revision_ids]
             other.inventory_store.prefetch(inventory_ids)
+
+        if pb is None:
+            pb = bzrlib.ui.ui_factory.progress_bar()
                 
         revisions = []
         needed_texts = set()
         i = 0
-        for rev_id in revision_ids:
-            i += 1
-            if progress:
-                progress.update('fetching revision', i, len(revision_ids))
-            rev = other.get_revision(rev_id)
+
+        failures = set()
+        for i, rev_id in enumerate(revision_ids):
+            pb.update('fetching revision', i+1, len(revision_ids))
+            try:
+                rev = other.get_revision(rev_id)
+            except bzrlib.errors.NoSuchRevision:
+                failures.add(rev_id)
+                continue
+
             revisions.append(rev)
             inv = other.get_inventory(str(rev.inventory_id))
             for key, entry in inv.iter_entries():
@@ -862,20 +858,24 @@
                 if entry.text_id not in self.text_store:
                     needed_texts.add(entry.text_id)
 
-        if progress:
-            progress.clear()
+        pb.clear()
                     
-        count = self.text_store.copy_multi(other.text_store, needed_texts)
+        count, cp_fail = self.text_store.copy_multi(other.text_store, 
+                                                    needed_texts)
         print "Added %d texts." % count 
         inventory_ids = [ f.inventory_id for f in revisions ]
-        count = self.inventory_store.copy_multi(other.inventory_store, 
-                                                inventory_ids)
+        count, cp_fail = self.inventory_store.copy_multi(other.inventory_store, 
+                                                         inventory_ids)
         print "Added %d inventories." % count 
         revision_ids = [ f.revision_id for f in revisions]
-        count = self.revision_store.copy_multi(other.revision_store, 
-                                               revision_ids)
-        return count
+
+        count, cp_fail = self.revision_store.copy_multi(other.revision_store, 
+                                                          revision_ids,
+                                                          permit_failure=True)
+        assert len(cp_fail) == 0 
+        return count, failures
        
+
     def commit(self, *args, **kw):
         from bzrlib.commit import commit
         commit(self, *args, **kw)

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-08-25 03:23:17 +0000
+++ bzrlib/errors.py	2005-08-25 06:14:17 +0000
@@ -99,3 +99,8 @@
         BzrCommandError.__init__(self, msg)
 
 
+class InstallFailed(BzrError):
+    def __init__(self, revisions):
+        self.revisions = revisions
+        msg = "Could not install revisions:\n%s" % " ,".join(revisions)
+        BzrError.__init__(self, msg)

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-08-24 12:13:13 +0000
+++ bzrlib/fetch.py	2005-08-25 06:14:17 +0000
@@ -15,23 +15,48 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 import bzrlib.errors
 from bzrlib.selftest.testrevision import make_branches
-from bzrlib.trace import mutter
+from bzrlib.trace import mutter, note
 from bzrlib.branch import Branch
+from bzrlib.progress import ProgressBar
 import sys
 import os
 
-def greedy_fetch(from_branch, to_branch, last_revision=None):
+def greedy_fetch(to_branch, from_branch, revision=None, pb=None):
+    """Copy a revision and all available ancestors from one branch to another
+    If no revision is specified, uses the last revision in the source branch's
+    revision history.
+    """
     from_history = from_branch.revision_history()
-    if last_revision is not None:
-        from_history = from_history[:from_history.index(last_revision)+1]
+    required_revisions = set(from_history)
+    all_failed = set()
+    if revision is not None:
+        required_revisions.add(revision)
+        try:
+            rev_index = from_history.index(revision)
+        except ValueError:
+            rev_index = None
+        if rev_index is not None:
+            from_history = from_history[:rev_index + 1]
+        else:
+            from_history = [revision]
     to_history = to_branch.revision_history()
     missing = []
     for rev_id in from_history:
         if not has_revision(to_branch, rev_id):
             missing.append(rev_id)
-
+    
+    count = 0
     while len(missing) > 0:
-        to_branch.install_revisions(from_branch, revision_ids=missing)
+        installed, failed = to_branch.install_revisions(from_branch, 
+                                                        revision_ids=missing,
+                                                        pb=pb)
+        count += installed
+        required_failed = failed.intersection(required_revisions)
+        if len(required_failed) > 0:
+            raise bzrlib.errors.InstallFailed(required_failed)
+        for rev_id in failed:
+            note("Failed to install %s" % rev_id)
+        all_failed.update(failed)
         new_missing = []
         for rev_id in missing:
             try:
@@ -45,9 +70,11 @@
                 if not has_revision(to_branch, parent):
                     new_missing.append(parent)
         missing = new_missing
+    return count, all_failed
 
 
 from testsweet import InTempDir
+from bzrlib.commit import commit
 def has_revision(branch, revision_id):
     try:
         branch.get_revision_xml(revision_id)
@@ -61,37 +88,63 @@
             os.mkdir(name)
             return Branch(name, init=True)
             
-        #highest indices a: 3, b: 4
+        #highest indices a: 5, b: 7
         br_a, br_b = make_branches()
         assert not has_revision(br_b, br_a.revision_history()[3])
         assert has_revision(br_b, br_a.revision_history()[2])
-        assert len(br_b.revision_history()) == 5
-        greedy_fetch(br_a, br_b, br_a.revision_history()[2])
+        assert len(br_b.revision_history()) == 7
+        assert greedy_fetch(br_b, br_a, br_a.revision_history()[2])[0] == 0
 
         # greedy_fetch is not supposed to alter the revision history
-        assert len(br_b.revision_history()) == 5
+        assert len(br_b.revision_history()) == 7
         assert not has_revision(br_b, br_a.revision_history()[3])
 
-        assert len(br_b.revision_history()) == 5
-        greedy_fetch(br_a, br_b, br_a.revision_history()[3])
+        assert len(br_b.revision_history()) == 7
+        assert greedy_fetch(br_b, br_a, br_a.revision_history()[3])[0] == 1
         assert has_revision(br_b, br_a.revision_history()[3])
         assert not has_revision(br_a, br_b.revision_history()[3])
         assert not has_revision(br_a, br_b.revision_history()[4])
-        greedy_fetch(br_b, br_a)
+
+        # When a non-branch ancestor is missing, it should be a failure, not
+        # exception
+        br_a4 = new_branch('br_a4')
+        count, failures = greedy_fetch(br_a4, br_a)
+        assert count == 6
+        assert failures == set((br_b.revision_history()[4],
+                                br_b.revision_history()[5])) 
+
+        assert greedy_fetch(br_a, br_b)[0] == 4
         assert has_revision(br_a, br_b.revision_history()[3])
         assert has_revision(br_a, br_b.revision_history()[4])
 
         br_b2 = new_branch('br_b2')
-        greedy_fetch(br_b, br_b2)
+        assert greedy_fetch(br_b2, br_b)[0] == 7
         assert has_revision(br_b2, br_b.revision_history()[4])
         assert has_revision(br_b2, br_a.revision_history()[2])
         assert not has_revision(br_b2, br_a.revision_history()[3])
 
         br_a2 = new_branch('br_a2')
-        greedy_fetch(br_a, br_a2)
+        assert greedy_fetch(br_a2, br_a)[0] == 9
         assert has_revision(br_a2, br_b.revision_history()[4])
         assert has_revision(br_a2, br_a.revision_history()[3])
 
+        br_a3 = new_branch('br_a3')
+        assert greedy_fetch(br_a3, br_a2)[0] == 0
+        for revno in range(4):
+            assert not has_revision(br_a3, br_a.revision_history()[revno])
+        assert greedy_fetch(br_a3, br_a2, br_a.revision_history()[2])[0] == 3
+        fetched = greedy_fetch(br_a3, br_a2, br_a.revision_history()[3])[0]
+        assert fetched == 3, "fetched %d instead of 3" % fetched
+        # InstallFailed should be raised if the branch is missing the revision
+        # that was requested.
+        self.assertRaises(bzrlib.errors.InstallFailed, greedy_fetch, br_a3,
+                          br_a2, 'pizza')
+        # InstallFailed should be raised if the branch is missing a revision
+        # from its own revision history
+        br_a2.append_revision('a-b-c')
+        self.assertRaises(bzrlib.errors.InstallFailed, greedy_fetch, br_a3,
+                          br_a2)
+
 
 
 if __name__ == '__main__':

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2005-08-25 03:23:17 +0000
+++ bzrlib/merge.py	2005-08-25 06:14:17 +0000
@@ -1,7 +1,24 @@
+# Copyright (C) 2005 Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+
 from bzrlib.merge_core import merge_flex, ApplyMerge3, BackupBeforeChange
 from bzrlib.changeset import generate_changeset, ExceptionConflictHandler
 from bzrlib.changeset import Inventory, Diff3Merge
-from bzrlib import find_branch
+from bzrlib.branch import find_branch
 import bzrlib.osutils
 from bzrlib.errors import BzrCommandError, UnrelatedBranches
 from bzrlib.delta import compare_trees
@@ -10,6 +27,7 @@
 import tempfile
 import shutil
 import errno
+from fetch import greedy_fetch
 
 
 # comments from abentley on irc: merge happens in two stages, each
@@ -110,18 +128,30 @@
         if not self.ignore_zero:
             print "%d conflicts encountered.\n" % self.conflicts
             
-def get_tree(treespec, temp_root, label):
+def get_tree(treespec, temp_root, label, local_branch=None):
     location, revno = treespec
     branch = find_branch(location)
     if revno is None:
+        revision = None
+    elif revno == -1:
+        revision = branch.last_patch()
+    else:
+        revision = branch.lookup_revision(revno)
+    return branch, get_revid_tree(branch, revision, temp_root, label,
+                                  local_branch)
+
+def get_revid_tree(branch, revision, temp_root, label, local_branch):
+    if revision is None:
         base_tree = branch.working_tree()
-    elif revno == -1:
-        base_tree = branch.basis_tree()
     else:
-        base_tree = branch.revision_tree(branch.lookup_revision(revno))
+        if local_branch is not None:
+            greedy_fetch(local_branch, branch, revision)
+            base_tree = local_branch.revision_tree(revision)
+        else:
+            base_tree = branch.revision_tree(revision)
     temp_path = os.path.join(temp_root, label)
     os.mkdir(temp_path)
-    return branch, MergeTree(base_tree, temp_path)
+    return MergeTree(base_tree, temp_path)
 
 
 def file_exists(tree, file_id):
@@ -196,29 +226,64 @@
     check_clean
         If true, this_dir must have no uncommitted changes before the
         merge begins.
+    all available ancestors of other_revision and base_revision are
+    automatically pulled into the branch.
     """
+    from bzrlib.revision import common_ancestor, MultipleRevisionSources
+    from bzrlib.errors import NoSuchRevision
     tempdir = tempfile.mkdtemp(prefix="bzr-")
     try:
         if this_dir is None:
             this_dir = '.'
         this_branch = find_branch(this_dir)
+        this_rev_id = this_branch.last_patch()
+        if this_rev_id is None:
+            raise BzrCommandError("This branch has no commits")
         if check_clean:
             changes = compare_trees(this_branch.working_tree(), 
                                     this_branch.basis_tree(), False)
             if changes.has_changed():
                 raise BzrCommandError("Working tree has uncommitted changes.")
-        other_branch, other_tree = get_tree(other_revision, tempdir, "other")
+        other_branch, other_tree = get_tree(other_revision, tempdir, "other",
+                                            this_branch)
+        if other_revision[1] == -1:
+            other_rev_id = other_branch.last_patch()
+            other_basis = other_rev_id
+        elif other_revision[1] is not None:
+            other_rev_id = other_branch.lookup_revision(other_revision[1])
+            other_basis = other_rev_id
+        else:
+            other_rev_id = None
+            other_basis = other_branch.last_patch()
         if base_revision == [None, None]:
             if other_revision[1] == -1:
                 o_revno = None
             else:
                 o_revno = other_revision[1]
-            base_revno = this_branch.common_ancestor(other_branch, 
-                                                     other_revno=o_revno)[0]
-            if base_revno is None:
                 raise UnrelatedBranches()
-            base_revision = ['.', base_revno]
-        base_branch, base_tree = get_tree(base_revision, tempdir, "base")
+            try:
+                base_revision = this_branch.get_revision(base_rev_id)
+                base_branch = this_branch
+            except NoSuchRevision:
+                base_branch = other_branch
+            base_tree = get_revid_tree(base_branch, base_rev_id, tempdir, 
+                                       "base")
+            base_is_ancestor = True
+        else:
+            base_branch, base_tree = get_tree(base_revision, tempdir, "base")
+            if base_revision[1] == -1:
+                base_rev_id = base_branch.last_patch()
+            elif base_revision[1] is None:
+                base_rev_id = None
+            else:
+                base_rev_id = base_branch.lookup_revision(base_revision[1])
+            if base_rev_id is not None:
+                base_is_ancestor = is_ancestor(this_rev_id, base_rev_id, 
+                                               MultipleRevisionSources(
+                                               this_branch, 
+                                               base_branch))
+            else:
+                base_is_ancestor = False
         if file_list is None:
             interesting_ids = None
         else:
@@ -238,6 +303,8 @@
         merge_inner(this_branch, other_tree, base_tree, tempdir, 
                     ignore_zero=ignore_zero, backup_files=backup_files, 
                     merge_type=merge_type, interesting_ids=interesting_ids)
+        if base_is_ancestor and other_rev_id is not None:
+            this_branch.add_pending_merge(other_rev_id)
     finally:
         shutil.rmtree(tempdir)
 

=== modified file 'bzrlib/revision.py'
--- bzrlib/revision.py	2005-08-05 19:23:12 +0000
+++ bzrlib/revision.py	2005-08-25 06:14:17 +0000
@@ -176,12 +176,19 @@
     behaves like Branch's.
     """
 
+    for ancestor_id, distance in iter_ancestors(revision_id, revision_source):
+        if ancestor_id == candidate_id:
+            return True
+    return False
+
+def iter_ancestors(revision_id, revision_source, only_present=False):
     ancestors = (revision_id,)
+    distance = 0
     while len(ancestors) > 0:
         new_ancestors = []
         for ancestor in ancestors:
-            if ancestor == candidate_id:
-                return True
+            if not only_present:
+                yield ancestor, distance
             try:
                 revision = revision_source.get_revision(ancestor)
             except bzrlib.errors.NoSuchRevision, e:
@@ -189,9 +196,67 @@
                     raise 
                 else:
                     continue
+            if only_present:
+                yield ancestor, distance
             new_ancestors.extend([p.revision_id for p in revision.parents])
         ancestors = new_ancestors
-
+        distance += 1
+
+
+def find_present_ancestors(revision_id, revision_source):
+    found_ancestors = {}
+    count = 0
+    anc_iter = enumerate(iter_ancestors(revision_id, revision_source,
+                         only_present=True))
+    for anc_order, (anc_id, anc_distance) in anc_iter:
+        if not found_ancestors.has_key(anc_id):
+            found_ancestors[anc_id] = (anc_order, anc_distance)
+    return found_ancestors
+    
+class AmbiguousBase(bzrlib.errors.BzrError):
+    def __init__(self, bases):
+        msg = "The correct base is unclear, becase %s are all equally close" %\
+            ", ".join(bases)
+        bzrlib.errors.BzrError.__init__(self, msg)
+        self.bases = bases
+
+def common_ancestor(revision_a, revision_b, revision_source):
+    """Find the ancestor common to both revisions that is closest to both.
+    """
+    from bzrlib.trace import mutter
+    a_ancestors = find_present_ancestors(revision_a, revision_source)
+    b_ancestors = find_present_ancestors(revision_b, revision_source)
+    a_intersection = []
+    b_intersection = []
+    # a_order is used as a tie-breaker when two equally-good bases are found
+    for revision, (a_order, a_distance) in a_ancestors.iteritems():
+        if b_ancestors.has_key(revision):
+            a_intersection.append((a_distance, a_order, revision))
+            b_intersection.append((b_ancestors[revision][1], a_order, revision))
+    mutter("a intersection: %r" % a_intersection)
+    mutter("b intersection: %r" % b_intersection)
+    def get_closest(intersection):
+        intersection.sort()
+        matches = [] 
+        for entry in intersection:
+            if entry[0] == intersection[0][0]:
+                matches.append(entry[2])
+        return matches
+
+    a_closest = get_closest(a_intersection)
+    if len(a_closest) == 0:
+        return None
+    b_closest = get_closest(b_intersection)
+    assert len(b_closest) != 0
+    mutter ("a_closest %r" % a_closest)
+    mutter ("b_closest %r" % b_closest)
+    if a_closest[0] in b_closest:
+        return a_closest[0]
+    elif b_closest[0] in a_closest:
+        return b_closest[0]
+    else:
+        raise AmbiguousBase((a_closest[0], b_closest[0]))
+    return a_closest[0]
 
 class MultipleRevisionSources(object):
     def __init__(self, *args):

=== modified file 'bzrlib/selftest/blackbox.py'
--- bzrlib/selftest/blackbox.py	2005-08-24 06:53:07 +0000
+++ bzrlib/selftest/blackbox.py	2005-08-25 06:14:17 +0000
@@ -347,3 +347,58 @@
         runbzr('log --forward')
 
         runbzr('info')
+
+
+
+def example_branch(test):
+    test.runbzr('init')
+
+    file('hello', 'wt').write('foo')
+    test.runbzr('add hello')
+    test.runbzr('commit -m setup hello')
+
+    file('goodbye', 'wt').write('baz')
+    test.runbzr('add goodbye')
+    test.runbzr('commit -m setup goodbye')
+
+
+class RevertCommand(ExternalBase):
+    def runTest(self):
+        example_branch(self)
+        file('hello', 'wt').write('bar')
+        file('goodbye', 'wt').write('qux')
+        self.runbzr('revert hello')
+        self.check_file_contents('hello', 'foo')
+        self.check_file_contents('goodbye', 'qux')
+        self.runbzr('revert')
+        self.check_file_contents('goodbye', 'baz')
+
+
+class MergeCommand(ExternalBase):
+    def runTest(self):
+        from bzrlib.branch import Branch
+        import os
+        os.mkdir('a')
+        os.chdir('a')
+        example_branch(self)
+        os.chdir('..')
+        self.runbzr('branch a b')
+        os.chdir('b')
+        file('goodbye', 'wt').write('quux')
+        self.runbzr(['commit',  '-m',  "more u's are always good"])
+
+        os.chdir('../a')
+        file('hello', 'wt').write('quuux')
+        # We can't merge when there are in-tree changes
+        self.runbzr('merge ../b', retcode=1)
+        self.runbzr(['commit', '-m', "Like an epidemic of u's"])
+        self.runbzr('merge ../b')
+        self.check_file_contents('goodbye', 'quux')
+        # Merging a branch pulls its revision into the tree
+        a = Branch('.')
+        b = Branch('../b')
+        a.get_revision_xml(b.last_patch())
+        print "Pending: %s" % a.pending_merges()
+#        assert a.pending_merges() == [b.last_patch()], "Assertion %s %s" \
+#        % (a.pending_merges(), b.last_patch())
+

=== modified file 'bzrlib/selftest/testbranch.py'
--- bzrlib/selftest/testbranch.py	2005-08-24 05:51:34 +0000
+++ bzrlib/selftest/testbranch.py	2005-08-25 06:14:17 +0000
@@ -26,3 +26,23 @@
         self.assertEquals(br.revision_history(), ["rev1",])
         br.append_revision("rev2", "rev3")
         self.assertEquals(br.revision_history(), ["rev1", "rev2", "rev3"])
+
+
+# TODO: rewrite this as a regular unittest, without relying on the displayed output        
+#         >>> from bzrlib.commit import commit
+#         >>> bzrlib.trace.silent = True
+#         >>> br1 = ScratchBranch(files=['foo', 'bar'])
+#         >>> br1.add('foo')
+#         >>> br1.add('bar')
+#         >>> commit(br1, "lala!", rev_id="REVISION-ID-1", verbose=False)
+#         >>> br2 = ScratchBranch()
+#         >>> br2.update_revisions(br1)
+#         Added 2 texts.
+#         Added 1 inventories.
+#         Added 1 revisions.
+#         >>> br2.revision_history()
+#         [u'REVISION-ID-1']
+#         >>> br2.update_revisions(br1)
+#         Added 0 revisions.
+#         >>> br1.text_store.total_size() == br2.text_store.total_size()
+#         True

=== modified file 'bzrlib/selftest/testrevision.py'
--- bzrlib/selftest/testrevision.py	2005-08-24 05:51:34 +0000
+++ bzrlib/selftest/testrevision.py	2005-08-25 06:14:17 +0000
@@ -20,21 +20,29 @@
 def make_branches():
     from bzrlib.branch import Branch
     from bzrlib.commit import commit
+    from bzrlib.revision import validate_revision_id
     import os
     os.mkdir("branch1")
     br1 = Branch("branch1", init=True)
-    commit(br1, "Commit one")
-    commit(br1, "Commit two")
-    commit(br1, "Commit three")
+    
+    commit(br1, "Commit one", rev_id="a@u-0-0")
+    commit(br1, "Commit two", rev_id="a@u-0-1")
+    commit(br1, "Commit three", rev_id="a@u-0-2")
 
     os.mkdir("branch2")
     br2 = Branch("branch2", init=True)
     br2.update_revisions(br1)
-    commit(br2, "Commit four")
-    commit(br2, "Commit five")
+    commit(br2, "Commit four", rev_id="b@u-0-3")
+    commit(br2, "Commit five", rev_id="b@u-0-4")
     revisions_2 = br2.revision_history()
     br1.add_pending_merge(revisions_2[4])
-    commit(br1, "Commit six")
+    commit(br1, "Commit six", rev_id="a@u-0-3")
+    commit(br1, "Commit seven", rev_id="a@u-0-4")
+    commit(br2, "Commit eight", rev_id="b@u-0-5")
+    br1.add_pending_merge(br2.revision_history()[5])
+    commit(br1, "Commit nine", rev_id="a@u-0-5")
+    br2.add_pending_merge(br1.revision_history()[4])
+    commit(br2, "Commit ten", rev_id="b@u-0-6")
     return br1, br2
 
 
@@ -59,3 +67,44 @@
         assert is_ancestor(revisions[3], revisions_2[4], br1)
         assert is_ancestor(revisions[3], revisions_2[3], sources)
         assert not is_ancestor(revisions[3], revisions_2[3], br1)
+
+
+class TestCommonAncestor(InTempDir):
+    """Test checking whether a revision is an ancestor of another revision"""
+    def runTest(self):
+        from bzrlib.revision import find_present_ancestors, common_ancestor
+        from bzrlib.revision import MultipleRevisionSources
+        br1, br2 = make_branches()
+        revisions = br1.revision_history()
+        revisions_2 = br2.revision_history()
+        sources = MultipleRevisionSources(br1, br2)
+
+        expected_ancestors_list = {revisions[3]:(0, 0), 
+                                   revisions[2]:(1, 1),
+                                   revisions_2[4]:(2, 1), 
+                                   revisions[1]:(3, 2),
+                                   revisions_2[3]:(4, 2),
+                                   revisions[0]:(5, 3) }
+        ancestors_list = find_present_ancestors(revisions[3], sources)
+        assert len(expected_ancestors_list) == len(ancestors_list)
+        for key, value in expected_ancestors_list.iteritems():
+            self.assertEqual(ancestors_list[key], value, 
+                              "key %r, %r != %r" % (key, ancestors_list[key],
+                                                    value))
+
+        self.assertEqual(common_ancestor(revisions[0], revisions[0], sources),
+                          revisions[0])
+        self.assertEqual(common_ancestor(revisions[1], revisions[2], sources),
+                          revisions[1])
+        self.assertEqual(common_ancestor(revisions[1], revisions[1], sources),
+                          revisions[1])
+        self.assertEqual(common_ancestor(revisions[2], revisions_2[4], sources),
+                          revisions[2])
+        self.assertEqual(common_ancestor(revisions[3], revisions_2[4], sources),
+                          revisions_2[4])
+        self.assertEqual(common_ancestor(revisions[4], revisions_2[5], sources),
+                          revisions_2[4])
+        self.assertEqual(common_ancestor(revisions[5], revisions_2[6], sources),
+                          revisions[4])
+        self.assertEqual(common_ancestor(revisions_2[6], revisions[5], sources),
+                          revisions_2[5])

=== modified file 'bzrlib/store.py'
--- bzrlib/store.py	2005-08-24 12:13:13 +0000
+++ bzrlib/store.py	2005-08-25 06:14:17 +0000
@@ -111,7 +111,7 @@
             af.close()
 
 
-    def copy_multi(self, other, ids):
+    def copy_multi(self, other, ids, permit_failure=False):
         """Copy texts for ids from other into self.
 
         If an id is present in self, it is skipped.  A count of copied
@@ -127,15 +127,25 @@
         for id in to_copy:
             count += 1
             pb.update('copy', count, len(to_copy))
-            self.add(other[id], id)
+            if not permit_failure:
+                self.add(other[id], id)
+            else:
+                try:
+                    entry = other[id]
+                except IndexError:
+                    failures.add(id)
+                    continue
+                self.add(entry, id)
+                
         assert count == len(to_copy)
         pb.clear()
         return count
 
 
-    def copy_multi_immutable(self, other, to_copy, pb):
+    def copy_multi_immutable(self, other, to_copy, pb, permit_failure=False):
         from shutil import copyfile
         count = 0
+        failed = set()
         for id in to_copy:
             p = self._path(id)
             other_p = other._path(id)
@@ -143,7 +153,16 @@
                 copyfile(other_p, p)
             except IOError, e:
                 if e.errno == errno.ENOENT:
-                    copyfile(other_p+".gz", p+".gz")
+                    if not permit_failure:
+                        copyfile(other_p+".gz", p+".gz")
+                    else:
+                        try:
+                            copyfile(other_p+".gz", p+".gz")
+                        except IOError, e:
+                            if e.errno == errno.ENOENT:
+                                failed.add(id)
+                            else:
+                                raise
                 else:
                     raise
             
@@ -151,7 +170,7 @@
             pb.update('copy', count, len(to_copy))
         assert count == len(to_copy)
         pb.clear()
-        return count
+        return count, failed
     
 
     def __contains__(self, fileid):

